use strict;
#------------------------------------------------------------------------------
# OAuth⥸塼
#						(C)2010 nabe / nabe@abk
#------------------------------------------------------------------------------
# HMAC-SHA1 
#
package Satsuki::Base::OAuth;
use Satsuki::Boolean;
our $VERSION = '0.20';
###############################################################################
# ܽ
###############################################################################
#------------------------------------------------------------------------------
# ڥ󥹥ȥ饯
#------------------------------------------------------------------------------
sub new {
	my $class = shift;
	my $self = {};		# 
	bless($self, $class);	# $self 򤳤Υ饹ȴϢդ

	my $ROBJ = shift;
	$self->{ROBJ} = $ROBJ;
	$self->{oauth_ver} = "1.0";
	$self->{signature_method} = 'HMAC-SHA1';
	return $self;
}

###############################################################################
# ᥤ롼
###############################################################################
#------------------------------------------------------------------------------
# ꥵФ˥ꥯȤ
#------------------------------------------------------------------------------
sub request_token {
	my $self = shift;
	my $h = shift;
	my $ROBJ = $self->{ROBJ};

	my $nonce = $self->generate_nonce();
	my $msg = "oauth_consumer_key=$h->{consumer_key}"
		. "&oauth_nonce=$nonce"
		. "&oauth_signature_method=$self->{signature_method}"
		. "&oauth_timestamp=$ROBJ->{TM}"
		. "&oauth_version=$self->{oauth_ver}"
		;
	my $sig = $self->generate_signature('GET', $h->{request_token_path}, $msg, $h->{consumer_secret});

	my %header = ('Authorization' => <<TEXT);
OAuth realm="$h->{realm}",
 oauth_consumer_key="$h->{consumer_key}",
 oauth_nonce="$nonce",
 oauth_signature="$sig",
 oauth_signature_method="$self->{signature_method}",
 oauth_timestamp="$ROBJ->{TM}",
 oauth_version="$self->{oauth_ver}"
TEXT

	my $http = $ROBJ->loadpm('Base::HTTP');
	$http->{error_to_root} = 1;
	my $res = $http->do_get($h->{request_token_path}, \%header);

	if (!ref($res)) { return undef; }	# error
	my $res = $self->parse_response( join('', @$res) );
	if ($res->{oauth_callback_confirmed} ne 'true') {
		$ROBJ->error("OAuth request_token response error\n");
		return undef;
	}
	return $res;
}

#------------------------------------------------------------------------------
# ǧQuery
#------------------------------------------------------------------------------
sub request_access_token {
	my $self = shift;
	my $h = shift;
	my $ROBJ = $self->{ROBJ};

	my $nonce = $self->generate_nonce();
	my $msg = "oauth_consumer_key=$h->{consumer_key}"
		. "&oauth_nonce=$nonce"
		. "&oauth_signature_method=$self->{signature_method}"
		. "&oauth_timestamp=$ROBJ->{TM}"
		. "&oauth_token=$h->{token}"
		. "&oauth_verifier=$h->{verifier}"
		. "&oauth_version=$self->{oauth_ver}"
		;
	my $sig = $self->generate_signature('GET', $h->{access_token_path}, $msg,
		$h->{consumer_secret}, $h->{token_secret});

	my %header = ('Authorization' => <<TEXT);
OAuth realm="$h->{realm}",
 oauth_consumer_key="$h->{consumer_key}",
 oauth_nonce="$nonce",
 oauth_signature="$sig",
 oauth_signature_method="$self->{signature_method}",
 oauth_timestamp="$ROBJ->{TM}",
 oauth_token="$h->{token}",
 oauth_verifier="$h->{verifier}",
 oauth_version="$self->{oauth_ver}"
TEXT
	my $http = $ROBJ->loadpm('Base::HTTP');
	$http->{error_to_root} = 1;
	my $res = $http->do_get($h->{access_token_path}, \%header);

	if (!ref($res)) { return undef; }	# error
	return $self->parse_response( join('', @$res) );
}

#------------------------------------------------------------------------------
# ǧQuery
#------------------------------------------------------------------------------
# status ΤбΤäĤ͡
sub status_update {
	my $self = shift;
	my ($h, $url, $req) = @_;
	my $ROBJ = $self->{ROBJ};

#  GET /photos?file=vacation.jpg&size=original HTTP/1.1
#  Host: photos.example.net
#  Authorization: OAuth realm="Photos",
#     oauth_consumer_key="dpf43f3p2l4k3l03",
#     oauth_token="nnch734d00sl2jdk",
#     oauth_signature_method="HMAC-SHA1",
#     oauth_timestamp="137131202",
#     oauth_nonce="chapoH",
#     oauth_signature="MdpQcU8iPSUjWoN%2FUDMsK2sui9I%3D"

	my $status = $req->{status};
	my $jcode = $ROBJ->loadpm('Jcode');	# ܸѴWrapper
	$status = $jcode->from_to(\$status, $ROBJ->{System_coding}, 'UTF-8');
	$req->{status} = $status;

	my $HTTP = $ROBJ->loadpm('Base::HTTP');
#	$HTTP->uri_encode($status);
	$self->urlencode_com($status);

	my $nonce = $self->generate_nonce();
	my $msg = "oauth_consumer_key=$h->{consumer_key}"
		. "&oauth_nonce=$nonce"
		. "&oauth_signature_method=$self->{signature_method}"
		. "&oauth_timestamp=$ROBJ->{TM}"
		. "&oauth_token=$h->{access_token}"
		. "&oauth_version=$self->{oauth_ver}"
		. "&status=$status"
		;
	my $sig = $self->generate_signature('POST', $url, $msg,
		$h->{consumer_secret}, $h->{access_token_secret});

	my %header = ('Authorization' => <<TEXT);
OAuth realm="$h->{realm}",
 oauth_consumer_key="$h->{consumer_key}",
 oauth_nonce="$nonce",
 oauth_signature="$sig",
 oauth_signature_method="$self->{signature_method}",
 oauth_timestamp="$ROBJ->{TM}",
 oauth_token="$h->{access_token}",
 oauth_version="$self->{oauth_ver}"
TEXT

	# $HTTP->{error_to_root} = 1;
	my $res = $HTTP->do_post($url, \%header, $req);

	if (!ref($res)) { return undef; }	# error
	return $self->parse_json( join('', @$res) );
}

###############################################################################
# ֥롼
###############################################################################
#------------------------------------------------------------------------------
# http://:xxx/ Υѡ
#------------------------------------------------------------------------------
sub parse_host_path {
	my $self = shift;
	my $url = shift;
	if ($url !~ m|^http://([^/:]*)(?::(\d+))?(.*)$|) {
		return undef;
	}
	return ($1, $3);
}

#------------------------------------------------------------------------------
# nonce
#------------------------------------------------------------------------------
sub generate_nonce {
	my ($self) = @_;
	my $salt = $self->{ROBJ}->{Salt64chars};
	my $nonce;
	my $base = $ENV{UNIQUE_ID} . $ENV{REMOTE_ADDR};
	for(my $i=0; $i<20; $i++) {
		$nonce .= substr($salt, int(rand(256+ord(substr($base,$i,1))*256) % 64), 1);
	}
	$nonce =~ s/[^\w\-]//g;
	return $nonce;
}

#------------------------------------------------------------------------------
# HMAC-SHA1
#------------------------------------------------------------------------------
sub hmac_sha1 {
	my $self = shift;
	my ($key, $msg) = @_;
	my $sha1;
	if ($Digest::SHA::PurePerl::VERSION) {
		$sha1 = Digest::SHA::PurePerl->new(1);
	} else {
		eval {
			require Digest::SHA1;
			$sha1 = Digest::SHA1->new;
		};
		if ($@) {
			require Digest::SHA::PurePerl;
			$sha1 = Digest::SHA::PurePerl->new(1);
		}
	}

	my $bs = 64;
	if (length($key) > $bs) {
		$key = $sha1->add($key)->digest;
		$sha1->reset;
	}
	my $k_opad = $key ^ ("\x5c" x $bs);
	my $k_ipad = $key ^ ("\x36" x $bs);
	$sha1->add($k_ipad);
	$sha1->add($msg);
	my $hk_ipad = $sha1->digest;
	$sha1->reset;
	$sha1->add($k_opad, $hk_ipad);

	my $b64d = $sha1->b64digest;
	$b64d = substr($b64d.'====', 0, ((length($b64d)+3)>>2)<<2);
	return $b64d;
}

#------------------------------------------------------------------------------
# ̾
#------------------------------------------------------------------------------
sub generate_signature {
	my $self = shift;
	my ($method, $url, $msg, $secret1, $secret2) = @_;

	# signature
	$self->urlencode($url, $msg);
	my $sig = $self->hmac_sha1("$secret1&$secret2", "$method&$url&$msg");
	$sig =~ s/=/%3D/g;
	return $sig;
}

#------------------------------------------------------------------------------
# URI󥳡
#------------------------------------------------------------------------------
sub urlencode {
	my $self = shift;
	foreach(@_) {
		$_ =~ s|([^a-zA-Z0-9\-\._\~])|
			my $x = '%' . unpack('H2', $1);
			$x =~ tr/a-f/A-F/;
			$x;
		|eg;
	}
}
sub urlencode_com {
	my $self = shift;
	foreach(@_) {
		$_ =~ s|([^\w!\(\)\*\-\.\~\/])|
			my $x = '%' . unpack('H2', $1);
			$x =~ tr/a-f/A-F/;
			$x;
		|eg;
	}
}

#------------------------------------------------------------------------------
# 쥹ݥ󥹥եΥѡ
#------------------------------------------------------------------------------
sub parse_response {
	my $self = shift;
	my @ary = split(/&/, shift);
	my %h;
	foreach (@ary) {
		my ($k, $v) = split(/=/, $_, 2);
		$v =~ tr/+/ /;
		$v =~ s/%([0-9a-fA-F][0-9a-fA-F])/chr(hex($1))/eg;
		$v =~ s/[\x00-\x08\x0b\x0c\x0e-\x1f\x7f]//g;	# TAB LF CRʳ楳ɤ
		$h{$k} = $v;
	}
	return \%h;
}

#------------------------------------------------------------------------------
# JSONΥѡΤΤѡ
#------------------------------------------------------------------------------
my %json_esc = (
'/' => '/',
'"' => '"',
"\\" => "\\",
"b" => "\x08",
"f" => "\x0c",
"n" => "\n",
"r" => "\r",
"t" => "\t"
);
my %json_val = (
'true'  => $Satsuki::Boolean::true,
'false' => $Satsuki::Boolean::false,
'null'  => undef
);
sub parse_json {
	my $self = shift;
	my $json = shift;
	my @str;
	# $self->{ROBJ}->debug($json);
	$json =~ s/[\x00-\x01]//g;
	$json =~ s/\\"/\x01/g;
	$json =~ s/"([^"]*)"/push(@str,$1),"\x00$#str\x00"/eg;
	# ʸν
	foreach(@str) {
		$_ =~ s|\\([/\\bfnrt])|$json_esc{$1}|g;
		$_ =~ s/\x01/"/g;
		$_ =~ s/\\u([0-9A-Za-z][0-9A-Za-z][0-9A-Za-z][0-9A-Za-z])/
			my $x = hex($1);
			if ($x < 2) { 
				"";
			} elsif ($x < 0x100) { 
				chr($x);
			} else {
				# UTF-16 to UTF-8
				  chr(0xE0 | (($x>>12) & 0x0f))
				. chr(0x80 | (($x>> 6) & 0x3f))
				. chr(0x80 | ( $x      & 0x3f));
			}
		/eg;
	}
	$json =~ s/\s*//g;
	#-------------------------------------------------------
	# JSON
	#-------------------------------------------------------
	my @ary;
	foreach(split(/,/, $json)) {
		while($_ =~ /^(.*?)([\[\]\{\}])(.*)$/) {
			if ($1 ne '') { push(@ary, $1); }
			push(@ary, $2);
			$_ = $3;
		}
		if ($_ ne '') { push(@ary, $_); }
	}
	my @stack;
	my $ret = (shift(@ary) eq '{') ? {} : [];
	my $c = $ret;
	eval {
	while(@ary) {
		my $v = shift(@ary);
		if ($v eq '}' || $v eq ']') {
			$c = pop(@stack);
			next;
		}
		# ϥåξ
		if (ref($c) eq 'HASH') {
			my $x = index($v, ':');
			my $key = substr($v, 0, $x);
			my $val = substr($v, $x+1);
			$key =~ s/^\x00(\d+)\x00$/$str[$1]/;
			if (exists $json_val{$val}) {
				$val = $json_val{$val};
			} else {
				$val =~ s/^\x00(\d+)\x00$/$str[$1]/;
			}
			if ($ary[0] eq '{') {
				shift(@ary);
				push(@stack, $c);
				$c = $c->{$key} = {};
			} elsif ($ary[0] eq '[') {
				shift(@ary);
				push(@stack, $c);
				$c = $c->{$key} = [];
			} else {
				$c->{$key} = $val;
			}
			next;
		}
		# ξ
		if ($v eq '{') {
			push(@stack, $c);
			push(@$c, {});
			$c = $c->[ $#$c ];
		} elsif ($v eq '[') {
			push(@stack, $c);
			push(@$c, []);
			$c = $c->[ $#$c ];
		} elsif (exists $json_val{$v}) {
			push(@$c, $json_val{$v});
		} else {
			$v =~ s/^\x00(\d+)\x00$/$str[$1]/;
			push(@$c, $v);
		}
	}
	}; # eval
	return $ret;
}

sub nest_hash_to_string {
	my $data = shift;
	my $sp   = shift || '';
	my $ret = '';
	if (ref($data) eq 'HASH') {
		foreach(sort(keys(%$data))) {
			my $k = $_;
			my $v = $data->{$_};
			if (ref($v) eq 'ARRAY') {
				$ret .= "$sp$k=[\n";
				$ret .= &nest_hash_to_string($v, "  $sp");
				$ret .= "$sp]\n";
				next;
			}
			if (ref($v) eq 'HASH') {
				$ret .= "$sp$k={\n";
				$ret .= &nest_hash_to_string($v, "  $sp");
				$ret .= "$sp}\n";
				next;
			}
			$ret .= "$sp$k=$v\n";
		}
	} else {
		foreach(@$data) {
			if (ref($_) eq 'ARRAY') {
				$ret .= "$sp\[\n";
				$ret .= &nest_hash_to_string($_, "  $sp");
				$ret .= "$sp]\n";
				next;
			}
			if (ref($_) eq 'HASH') {
				$ret .= "$sp\{\n";
				$ret .= &nest_hash_to_string($_, "  $sp");
				$ret .= "$sp}\n";
				next;
			}
			$ret .= "$sp$_\n";
		}
	}
	return $ret;
}



###############################################################################
# 顼
###############################################################################
sub debug {
	my $self = shift;
	$self->{ROBJ}->debug(@_);	# debug-safe
}

1;
